/***************************************************************************
 *   Copyright (C) 2005 by Jeff Ferr                                       *
 *   root@sat                                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 a***************************************************************************/
#ifndef J_HTTP_H
#define J_HTTP_H

#include "jobject.h"

#include <string>
#include <vector>

#include <stdint.h>
#include <string.h>

#define MAXBUFFER	8192 

namespace jcommon {

/** 
 * \brief Enumerao para saber o mtodo de uma requisio HTTP.
 * 
 */
enum HTTPTipo {
	GET,
	POST,
	HEAD,
	OPTIONS,
	ERRO
};

/** \brief Enumerao de possveis tokens que pode ser retornados 
 * pelo analizador lxico HTTPlexer. 
 *
 */
enum Token {
	ERROR = -1,
	CRLF = 256,
	LWS,
	TOKEN,
	STRING,
	QUERYVAL,
	HEADERVAL,
	EOB,
	SP = ' ',
	WT = '\t',
	CR = '\r',
	LF = '\n'
};

/** 
 * \brief Estrutura interna para abstrao dos cabealhos e 
 * query strings de uma requisio HTTP. 
 * 
 */
struct Header {
	
	/** \brief Nome do cabealho ou query string. */
	char nome[256];
	/** \brief Valor do cabealho ou query string. */
	char valor[256];
	
	/** 
	 * \brief Operador de igualdade.
	 *
	 */
	bool operator==(Header &h) 
	{
		return !strcmp(h.nome,nome) && !strcmp(valor,h.valor);
	}
	
	/** 
	 * \brief Operador de igualdade para comparao do nome do cabealho. 
	 *
	 */
	bool operator==(char *h) 
	{
		return !strcmp(nome,h);
	}
	
	Header() 
	{
		nome[0] = 0;
		valor[0] = 0;
	}
};

/** 
 * \brief Estrutura interna para representar uma requisio HTTP.
 *
 * Ela e usada pela classe HTTP para encapsular os dados contidos em 
 * uma requisio. 
 *
 */
struct Request {
	/** \brief Metodo da requisio. */
	HTTPTipo metodo;
	/** \brief Usado para colocar o codigo de resposta da requisico. */
	int codigo;
	/** \brief Verso HTTP do cliente 1.x onde x e a versao. */
	char versao;
	/** \brief Arquivo solicitado pelo cliente. */
	char arquivo[1024];
	/** \brief Vetor de query strings passados na requisico. */
	std::vector<Header> query;
	/** \brief Vetor de cabealhos contidos na requisico do
	 * cliente. */
	std::vector<Header *> cabecalho;
	/** \brief Caso o metodo usado seja o POST, aqui sera colocado os dados passados no corpo da requisio. */
	char * body;
	
	Request() 
	{
		codigo = 0; //se td OK retorna 200
		versao = '1';
		metodo = ERRO;
		arquivo[0] = 0;
		body = NULL;
	}

	~Request() {
	}
};

/** 
 * \brief Classe que implementa um analisador lxico para o protocolo HTTP.
 *
 * Esta classe abstrai blocos de caracteres contidos no buffer em Tokens 
 * que sero analisados pelo parser HTTP implementado na classe HTTP.
 *
 */
class HTTPlexer {
	
	private:
		/** \brief */
		char * buffer;
		/** \brief */
		size_t bufsize;
		/** \brief */
		size_t cur;
		/** \brief */
		size_t pos;

	public:
		/**
		 * \brief Construtor da classe.
		 * 
		 * \param buf  o buffer de recepo a ser analisado. 
		 * \param tam  o tamanho em bytes do buffer.
		 * 
		 */
		HTTPlexer(char *buf, int tam);

		/**
		 * \brief Destructor.
		 *
		 */
		~HTTPlexer();
		
		/**  \brief Mtodo que analisa o buffer byte a byte para abstrair 
		 * a coleta de tokens.
		 * 
		 * \return O proximo caractere do buffer ou EOB em caso de fim de buffer.
		 *
		 */
		int GetChar();
		
		/** 
		 * \brief Mtodo para voltar a posio do caractere atual do buffer em 
		 * uma posio. 
		 *
		 */
		void Unget();
		
		/** 
		 * \brief Mtodo para retornar o proximo token encontrado a partir da
		 * posio atual do buffer, atualizando-a para depois do token.
		 * 
		 * \return Cdigo do token encontrado. Ou o prprio caractere se no 
		 * for encontrado nenhum token para a entrada.
		 *
		 */
		int GetToken();
		
		/** 
		 * \brief Mtodo estatico para tratar os caracteres de escape que podem
		 * vir codificados na url.
		 *
		 */
		static int decode(char * cod,char *dec);
		
		/** 
		 * \brief Mtodo para testar se um caratere  um separador de tokens HTTP ou no.
		 * 
		 * \param c  o caractere a ser comparado.
		 * 
		 * \return <i>true</i> caso o caractere c seja um separador HTTP. Ou false
		 * caso contrario.
		 *
		 */
		bool is_separator(char c);	
		
		/** 
		 * \brief Mtodo que retorna o proximo token retornando o lexeme do 
		 * token encontrado.
		 * 
		 * \param t  o codigo do token achado.
		 * \param len  o tamanho do lexeme achado.
		 * 
		 * \return A posio de inicio do lexeme.
		 *
		 */
		char * GetNextToken(int &t,int &len);
		
		/** 
		 * \brief Mtodo especial para recuperar o token HEADERVAL que possui 
		 * separadores diferentes que os definidos no HTTP (CRLF, LF ou EOB).
		 * 
		 * \return O cdigo do token achado.
		 *
		 */
		int GetHeaderVal();
		
		/** \brief Mtodo especial para recuperar o token QUERYVAL que possui 
		 * separadores diferentes que os definidos no HTTP (S, CRLF LF ou '&').
		 * 
		 * \return O cdigo do token achado.
		 *
		 */
		int GetQueryVal();
		
		/** 
		 * \brief Mtodo para testar se um caratere  um caractere de controle (0 a 32 e o 127).
		 * 
		 * \param c  o caractere a ser comparado.
		 * 
		 * \return true caso o caractere c seja um caractere de controle. Ou false
		 * caso contrario.
		 *
		 */
		bool is_ctl(char c);

		/** 
		 * \brief Mtodo para testar se um caratere  um digito HEXA.
		 * 
		 * \param c  o caractere a ser comparado.
		 * 
		 * \return true caso o caractere c seja um digito HEXA. Ou false
		 * caso contrario.
		 *
		 */
		static bool is_hex(char c);
		
		/** 
		 * \brief Mtodo para testar se o caractere atual  o fim de buffer.
		 * 
		 * \return true caso o buffer tenha Acabado Ou false caso contrario.
		 *
		 */
		bool is_eob();
		
		/** 
		 * \brief Mtodo para recuperar o lexeme do ultimo token reconhecido pelo 
		 * analisador lxico.
		 * 
		 * \param len  o tamanho em bytes do lexeme.
		 * 
		 * \return A posio inicial do lexeme.
		 *
		 */
		char * GetToken(int &len);

};

/**
 * \brief Arquivo de cabealho das classes e estruturas
 * para manipulao de requisies HTTP.
 * Nela esto todos os mtodos necessrios para tratar
 * uma requisio HTTP de forma transparente.
 * 
 * \author Tiago Braw
 */
class HTTP : public jcommon::Object{

	private:
		/** \brief */
		bool fim;
		/** \brief */
		Request r;
		/** \brief */
		int buflen;
		/** \brief */
		char buffer[MAXBUFFER];
		
	protected:
		/** \brief Mtodo que faz a anlise sintatica do buffer, atravs dos
		 * Tokens retornados pelo analisador lxico HTTPlexer.
		 * 
		 * \return true se a anlise sintatica foi terminada com sucesso. Ou false
		 * caso contrrio.
		 * 
		 * \todo Analise para, se uma requisio HTTP for do tipo POST, verifique 
		 * se o cliente enviou todos os dados corretamente.
		 *
		 */
		bool Parsear();
		
	public:
		/**
		 * \brief Constructor.
		 *
		 */
		HTTP();

		/**
		 * \brief Destructor.
		 *
		 */
		virtual ~HTTP();

		/** 
		 * \brief Mtodo para saber se uma requisio chegou a seu fim, ou no. 
		 * 
		 * \return true se a requisio chegou ao seu final. Ou false caso contrario.
		 *
		 */
		bool Fim();

		/** 
		 * \brief Mtodo para adcionar dados ao buffer de requisio. 
		 * 
		 * \param add  a string a ser adicionada ao buffer.
		 *
		 */
		void Add(char * add, int tam);
		
		/** 
		 * \brief Mtodo para recuperar a string de erro HTTP indexada pelo 
		 * codigo de requisio.
		 * 
		 * \param cod  o cdigo da requisio.
		 * 
		 * \return Uma string descrevendo o evento, ou "Internal Server Error" caso 
		 * o evento no exista. 
		 *
		 */
		static std::string Status(int cod);
		
		/**
		 * \brief Mtodo para avaliar o MIME type do arquivo da requisio.
		 * 
		 * \return O mimetype do arquivo requisitado.
		 *
		 */
    std::string GetMIME();
		
		/** 
		 * \brief Mtodo para procurar por valores de cabealhos contidos na requisio.
		 * 
		 * \param header  o cabealho a ser procurado.
		 * 
		 * \return O valor do cabealho encontrado, ou NULL caso no exista.
		 *
		 */
		char * GetHeader(const char * header);

		/** 
		 * \brief Mtodo para procurar por valores de querys contidos na requisio.
		 * 
		 * \param header  o query a ser procurado.
		 * 
		 * \return O valor do query encontrado, ou NULL caso no exista.
		 *
		 */
		char * GetQuery(const char * query);
		
		/** 
		 * \brief Mtodo para retorno do ponteiro para a estrutura Request montada.
		 * 
		 * \return O ponteiro para estrutura interna Request montada.
		 *
		 */
		Request * GetRequest() 
		{ 
			return &r; 
		}
		
		/** 
		 * \brief Mtodo para limpar a estrutura Request e o buffer para preparar 
		 * o objeto para parsear outra requisio HTTP.
		 *
		 */
		void Clear();
		
		/** 
		 * \brief Mtodo que retorna o buffer de recepcao puro... S para efeito 
		 * de debugging.
		 * 
		 * \return O buffer.
		 *
		 */
		char * GetBuffer() 
		{ 
			return buffer; 
		}

};

}

#endif

