/***************************************************************************
 *   Copyright (C) 2005 by Jeff Ferr                                       *
 *   root@sat                                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 a***************************************************************************/
#ifndef J_JSON_H
#define J_JSON_H

#ifdef WIN32
#define snprintf _snprintf
#define strncasecmp strnicmp
#define strcasecmp  stricmp
#endif

#include "jobject.h"

#include <stdio.h>

namespace jcommon {

/**
 * \brief
 * 
 * \author Jeff Ferr
 */
class JSONNode;
class JSONPullEvent;
class JSONPullEventQueue;
class JSONReader;
class JSONObjectNode;
class JSONArrayNode;
class JSONPullParser;
class JSONArrayList;
class JSONQueue;
class JSONStringBuffer;

typedef struct tagJSONArrayListNode JSONArrayListNode_t;

class JSONArrayList {
public:
	static const int LAST_INDEX;

	JSONArrayList( int initCount = 2 );
	virtual ~JSONArrayList();

	int getCount() const;
	int append( void * value );
	const void * getItem( int index ) const;
	void * takeItem( int index );
	void sort( int ( * cmpFunc )( const void *, const void * ) );

private:
	JSONArrayList( JSONArrayList & );
	JSONArrayList & operator=( JSONArrayList & );

	int mMaxCount;
	int mCount;
	void ** mFirst;
};

class JSONQueue {
public:
	JSONQueue();
	virtual ~JSONQueue();

	void push( void * item );
	void * pop();
	void * top();

private:
	void ** mEntries;
	unsigned int mHead;
	unsigned int mTail;
	unsigned int mCount;
	unsigned int mMaxCount;
};

class JSONStringBuffer {
public:
	JSONStringBuffer();
	virtual ~JSONStringBuffer();
	int append( char c );
	int append( const char * value, int size = 0 );
	int getSize() const;
	const char * getBuffer() const;
	char * takeBuffer();
	void clean();

private:
	JSONStringBuffer( JSONStringBuffer & );
	JSONStringBuffer & operator=( JSONStringBuffer & );

	void init();

	char * mBuffer;
	int mMaxSize;
	int mSize;	
};

class JSONReader {
public:
	/**
	 * @param  parser : act as reader's context
	 * @param  c : a char in json stream
	 */
	virtual void read( JSONPullParser * parser, char c ) = 0;

	/**
	 * convert internal json string to event
	 * @return NULL : this reader cannot generate any event or error occured
	 */
	virtual JSONPullEvent * getEvent( JSONPullParser * parser ) = 0;

protected:
	JSONStringBuffer * mBuffer;

	JSONReader();
	virtual ~JSONReader();

	friend class JSONPullParser;

	/// help to call parser->changeReader
	void changeReader( JSONPullParser * parser, JSONReader * reader );

	/// help to call parser->setError
	static void setError( JSONPullParser * parser, const char * error );

	/// help to call parser->isWait4Name
	int isWait4Name( JSONPullParser * parser );

	/// help to call parser->isObjectParent
	int isObjectParent( JSONPullParser * parser );

private:
	JSONReader( JSONReader & );
	JSONReader & operator=( JSONReader & );
};

class JSONSpaceReader : public JSONReader {
public:
	JSONSpaceReader( int hasReadComma = 0 );
	virtual ~JSONSpaceReader();
	virtual void read( JSONPullParser * parser, char c );
	virtual JSONPullEvent * getEvent( JSONPullParser * parser );

private:
	int mHasReadComma;
};

class JSONStartObjectReader : public JSONReader {
public:
	JSONStartObjectReader( int hasReadLeftBrace = 0 );
	virtual ~JSONStartObjectReader();
	virtual void read( JSONPullParser * parser, char c );
	virtual JSONPullEvent * getEvent( JSONPullParser * parser );

private:
	int mHasReadLeftBrace;
};

class JSONEndObjectReader : public JSONReader {
public:
	JSONEndObjectReader();
	virtual ~JSONEndObjectReader();
	virtual void read( JSONPullParser * parser, char c );
	virtual JSONPullEvent * getEvent( JSONPullParser * parser );
};

class JSONStartArrayReader : public JSONReader {
public:
	JSONStartArrayReader();
	virtual ~JSONStartArrayReader();
	virtual void read( JSONPullParser * parser, char c );
	virtual JSONPullEvent * getEvent( JSONPullParser * parser );

private:
	int mHasReadLeftBracket;
};

class JSONEndArrayReader : public JSONReader {
public:
	JSONEndArrayReader();
	virtual ~JSONEndArrayReader();
	virtual void read( JSONPullParser * parser, char c );
	virtual JSONPullEvent * getEvent( JSONPullParser * parser );
};

class JSONNameReader : public JSONReader {
public:
	JSONNameReader();
	virtual ~JSONNameReader();
	virtual void read( JSONPullParser * parser, char c );
	virtual JSONPullEvent * getEvent( JSONPullParser * parser );

private:
	char mLastChar;
	int mHasReadStartQuot;
};

class JSONValueReader : public JSONReader {
public:
	JSONValueReader();
	virtual ~JSONValueReader();
	virtual void read( JSONPullParser * parser, char c );
	virtual JSONPullEvent * getEvent( JSONPullParser * parser );
};

class JSONStringReader : public JSONReader {
public:
	JSONStringReader();
	virtual ~JSONStringReader();
	virtual void read( JSONPullParser * parser, char c );
	virtual JSONPullEvent * getEvent( JSONPullParser * parser );

private:
	char mLastChar;
	int mHasReadStartQuot;
};

class JSONNumberReader : public JSONReader {
public:
	JSONNumberReader();
	virtual ~JSONNumberReader();
	virtual void read( JSONPullParser * parser, char c );
	virtual JSONPullEvent * getEvent( JSONPullParser * parser );
};

class JSONBooleanReader : public JSONReader {
public:
	JSONBooleanReader();
	virtual ~JSONBooleanReader();
	virtual void read( JSONPullParser * parser, char c );
	virtual JSONPullEvent * getEvent( JSONPullParser * parser );
};

class JSONNullReader : public JSONReader {
public:
	JSONNullReader();
	virtual ~JSONNullReader();
	virtual void read( JSONPullParser * parser, char c );
	virtual JSONPullEvent * getEvent( JSONPullParser * parser );
};

class JSONCommentReader : public JSONReader {
public:
	JSONCommentReader( JSONReader * savedReader );
	virtual ~JSONCommentReader();
	virtual void read( JSONPullParser * parser, char c );
	virtual JSONPullEvent * getEvent( JSONPullParser * parser );

private:
	int mHasReadStartSolidus;
	JSONReader * mSavedReader;
};

class JSONPaddingReader : public JSONReader {
public:
	JSONPaddingReader();
	virtual ~JSONPaddingReader();
	virtual void read( JSONPullParser * parser, char c );
	virtual JSONPullEvent * getEvent( JSONPullParser * parser );
};

class JSONNode {
public:
	enum { eObject, ePair, eArray, eString, eDouble, eInt, eBoolean, eNull, eComment };

	JSONNode( int type );
	virtual ~JSONNode();

	int getType() const;
	JSONNode * getParent() const;
	void setParent( JSONNode * parent );

	int isLastChild() const;

protected:
	const int mType;
	JSONNode * mParent;
};

class JSONStringNode : public JSONNode {
public:
	JSONStringNode( const char * value );
	virtual ~JSONStringNode();

	const char * getValue();

private:
	char * mValue;
};

class JSONDoubleNode : public JSONNode {
public:
	JSONDoubleNode( double value );
	virtual ~JSONDoubleNode();

	double getValue();

private:
	double mValue;
};

class JSONIntNode : public JSONNode {
public:
	JSONIntNode( int value );
	virtual ~JSONIntNode();

	int getValue();

private:
	int mValue;
};

class JSONNullNode : public JSONNode {
public:
	JSONNullNode();
	virtual ~JSONNullNode();
};

class JSONBooleanNode : public JSONNode {
public:
	JSONBooleanNode( bool value );
	virtual ~JSONBooleanNode();

	bool getValue();

private:
	bool mValue;
};

class JSONCommentNode : public JSONNode {
public:
	JSONCommentNode( const char * comment );
	virtual ~JSONCommentNode();

	const char * getValue();

private:
	char * mValue;
};

class JSONPairNode : public JSONNode {
public:
	JSONPairNode();
	virtual ~JSONPairNode();

	void setName( const char * name );
	const char * getName();

	void setValue( JSONNode * value );
	JSONNode * getValue();

private:
	char * mName;
	JSONNode * mValue;
};

class JSONObjectNode : public JSONNode {
public:
	JSONObjectNode();
	virtual ~JSONObjectNode();

	int getCount();

	int addValue( JSONPairNode * value );
	JSONPairNode * getValue( int index );
	JSONPairNode * takeValue( int index );

	int Find( const char * name );

	int isLastChild( const JSONNode * value ) const;

private:
	JSONArrayList * mValueList;
};

class JSONArrayNode : public JSONNode {
public:
	JSONArrayNode();
	virtual ~JSONArrayNode();

	int getCount();

	int addValue( JSONNode * value );
	JSONNode * getValue( int index );
	JSONNode * takeValue( int index );

	int isLastChild( const JSONNode * value ) const;

private:
	JSONArrayList * mValueList;
};

class JSONPullParser {
public:
	JSONPullParser();
	~JSONPullParser();

	/// append more input json source
	int append( const char * source, int len );

	/// @return NOT NULL : the pull event
	/// @return NULL : error or need more input
	JSONPullEvent * getNext();	

	/// @return NOT NULL : the detail error message
	/// @return NULL : no error
	const char * getError();

	int getLevel();

protected:
	void changeReader( JSONReader * reader );

	void setError( const char * error );

	int isWait4Name();

	int isObjectParent();

	friend class JSONReader;

private:
	JSONPullEventQueue * mEventQueue;
	JSONReader * mReader;
	JSONArrayList * mStack;

	JSONArrayList * mNameList;

	int mLevel;

	char * mError;

	char mErrorSegment[ 32 ];
	int mErrorIndex;
	int mColIndex, mRowIndex;

	int mObjectCount;
};

class JSONDomParser {
public:
	JSONDomParser();
	~JSONDomParser();

	int append( const char * source, int len );

	const char * getError() const;

	const JSONNode * getValue() const;

private:

	void buildTree();

	void addNode( JSONNode * node );

	JSONPullParser * mParser;
	JSONArrayNode * mContainer;
	JSONNode * mCurrent;

private:
	// disable these two default member functions
	JSONDomParser( JSONDomParser & );
	JSONDomParser & operator=( JSONDomParser & );
};

class JSONDomBuffer {
public:
	JSONDomBuffer( const JSONNode * node, int indent = 1 );
	~JSONDomBuffer();

	const char * getBuffer() const;
	int getSize() const;

private:
	JSONStringBuffer * mBuffer;

	static void dump( const JSONNode * node,
			JSONStringBuffer * buffer, int level );
	static void dumpObject( const JSONNode * node,
			JSONStringBuffer * buffer, int level );
	static void dumpArray( const JSONNode * node,
			JSONStringBuffer * buffer, int level );
};

class JSONPullEvent {
public:
	enum EventType { eStartObject, eEndObject, eStartArray, eEndArray, eNull,
			eName, eString, eNumber, eBoolean, eComment, ePadding };

	JSONPullEvent( int eventType );
	virtual ~JSONPullEvent();

	int getEventType();

private:
	/// Private copy constructor and copy assignment ensure classes derived from
	/// this cannot be copied.
	JSONPullEvent( JSONPullEvent & );
	JSONPullEvent & operator=( JSONPullEvent & );

protected:
	const int mEventType;
};

class JSONPullEventQueue {
public:
	JSONPullEventQueue();
	~JSONPullEventQueue();

	void enqueue( JSONPullEvent * event );
	JSONPullEvent * dequeue();

private:
	JSONPullEventQueue( JSONPullEventQueue & );
	JSONPullEventQueue & operator=( JSONPullEventQueue & );

	JSONQueue * mQueue;
};

class JSONStartObjectEvent : public JSONPullEvent {
public:
	JSONStartObjectEvent();
	virtual ~JSONStartObjectEvent();
};

class JSONEndObjectEvent : public JSONPullEvent {
public:
	JSONEndObjectEvent();
	virtual ~JSONEndObjectEvent();
};

class JSONStartArrayEvent : public JSONPullEvent {
public:
	JSONStartArrayEvent();
	virtual ~JSONStartArrayEvent();
};

class JSONEndArrayEvent : public JSONPullEvent {
public:
	JSONEndArrayEvent();
	virtual ~JSONEndArrayEvent();
};

class JSONTextEvent : public JSONPullEvent {
public:
	JSONTextEvent( int eventType );
	virtual ~JSONTextEvent();

	void setText( const char * text, int len );
	const char * getText() const;

private:
	char * mText;
};

class JSONNullEvent : public JSONTextEvent {
public:
	JSONNullEvent();
	virtual ~JSONNullEvent();
};

class JSONNameEvent : public JSONTextEvent {
public:
	JSONNameEvent();
	virtual ~JSONNameEvent();
};

class JSONStringEvent : public JSONTextEvent {
public:
	JSONStringEvent();
	virtual ~JSONStringEvent();
};

class JSONNumberEvent : public JSONTextEvent {
public:
	JSONNumberEvent();
	virtual ~JSONNumberEvent();
};

class JSONBooleanEvent : public JSONTextEvent {
public:
	JSONBooleanEvent();
	virtual ~JSONBooleanEvent();
};

class JSONCommentEvent : public JSONTextEvent {
public:
	JSONCommentEvent();
	virtual ~JSONCommentEvent();
};

class JSONPaddingEvent : public JSONTextEvent {
public:
	JSONPaddingEvent();
	virtual ~JSONPaddingEvent();
};

class JSONCodec {
public:
	static int decode( const char * encodeValue, JSONStringBuffer * outBuffer );

	static int encode( const char * decodeValue, JSONStringBuffer * outBuffer );

	static int encode( double value, JSONStringBuffer * outBuffer );

	static const char ESCAPE_CHARS [];
	static const char UNESCAPED_CHARS [];

private:
	JSONCodec();
};


/// DFS iterator -- Depth First Searc
class JSONIterator {
public:
	/// node as tree node, iterator this tree by DFS
	JSONIterator( const JSONNode * node );
	~JSONIterator();

	/// @return NULL : reach the end
	const JSONNode * getNext();

private:
	const JSONNode * mRoot;
	const JSONNode * mCurrent;

private:
	JSONIterator( JSONIterator & );
	JSONIterator & operator=( JSONIterator & );
};

}

#endif
