/***************************************************************************
 *   Copyright (C) 2005 by Jeff Ferr                                       *
 *   root@sat                                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "jdatagramsocket.h"
#include "jsocketexception.h"
#include "jsockettimeoutexception.h"
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

namespace jsocket {

uint16_t DatagramSocket::_used_port = 1024;

DatagramSocket::DatagramSocket(std::string host_, uint16_t port_)
{
    _local = NULL;
    _address = NULL;
    _buffer = NULL;
    _sock_opt = NULL;
    _is = NULL;
    _os = NULL;

	CreateSocket();
	ConnectSocket(InetAddress::GetByName(host_), port_);
	InitStream();

	_sent_bytes = 0;
	_receive_bytes = 0;
}

DatagramSocket::DatagramSocket(uint16_t port_)
{
    _local = NULL;
    _address = NULL;
    _buffer = NULL;
    _sock_opt = NULL;
    _is = NULL;
    _os = NULL;

	CreateSocket();
	InetAddress *address = InetAddress::GetByName("localhost");

	if (port_ == 0) {
		while(true) {
			try {
				BindSocket(address, ++_used_port);
			} catch(SocketException &e) {
				continue;
			}

			break;
		}
	} else {
		BindSocket(_address, port_);
	}

	InitStream();

	_sent_bytes = 0;
	_receive_bytes = 0;
}

DatagramSocket::~DatagramSocket()
{
	try {
		Close();
	} catch (...) {
	}

    if (_local) {
        delete _local;
    }

    if (_address) {
        delete _address;
    }

    if (_buffer) {
        delete _buffer;
    }

    if (_sock_opt) {
        delete _sock_opt;
    }

    if (_is) {
        delete _is;
    }

    if (_os) {
        delete _os;
    }
}

/** Private */

void DatagramSocket::CreateSocket()
{
	_fd = ::socket(AF_INET, SOCK_DGRAM,IPPROTO_UDP);
	
	if (_fd < 0) {
		throw SocketException("Create datagram socket error !");
	}
}

void DatagramSocket::BindSocket(InetAddress *addr_, uint16_t local_port_)
{
	_local = addr_;
   
	bzero(&_lsock, sizeof(_lsock));
   
	_lsock.sin_family = AF_INET;
	_lsock.sin_addr.s_addr = htonl(INADDR_ANY);
   
	if(local_port_ != 0) {
		_lsock.sin_port = htons(local_port_);
	} else {
		int p = -1;

		_lsock.sin_port = htons(p);
	}
   
	if (::bind(_fd, (SA *)&_lsock, sizeof(_lsock)) < 0) {
		throw SocketException("Bind datagram socket error !");
	}
}

void DatagramSocket::ConnectSocket(InetAddress *addr_, uint16_t port_)
{
	_address = addr_;
	
	bzero(&_server_sock, sizeof(_server_sock));
	
	_server_sock.sin_family = AF_INET;
	
	if(_address == NULL) {
		_server_sock.sin_addr.s_addr = htonl(INADDR_ANY);
	} else {
		_server_sock.sin_addr.s_addr = ::inet_addr(_address->GetHostAddress().c_str());
	}
	
	_server_sock.sin_port = htons(port_);
}

void DatagramSocket::InitStream()
{
	_buffer = new DatagramSocketBuffer(_fd, RemoteAddress());
    
	_is = new std::istream(_buffer); // dsb
	_os = new std::ostream(_buffer);
}

/** End */

std::istream & DatagramSocket::GetInputStream()
{
	return *_is;
}

std::ostream & DatagramSocket::GetOutputStream()
{
	return *_os;
}

int DatagramSocket::Receive(char *data_, int size_)
{
	int n, length = sizeof(_server_sock);

	n = ::recvfrom(_fd, data_, size_, 0, (SA *)&_server_sock, (socklen_t*)&length);
	
	if (n < 0 && errno == EAGAIN) {
		throw SocketTimeoutException("Socket receive timeout exception !");
	} else if (n < 0) {
		throw SocketException("Read socket error !");
	}

	_receive_bytes += n;

    return n;
}

int DatagramSocket::Send(char *data_, int size_, bool notify_)
{
	int n;

	if (notify_ == false) {	
		n = ::sendto(_fd, data_, size_, 0, (SA *)&_server_sock, sizeof(_server_sock));
	} else {
		n = ::send(_fd, data_, size_, 0);
	}
	
	if (n < 0 && errno == EAGAIN) {
		throw SocketTimeoutException("Socket send timeout exception !");
	} else if (n < 0) {
		throw SocketException("Write udp data error !");
	}

	_sent_bytes += n;
	
	return n;
}

void DatagramSocket::Close()
{
	if (close(_fd) < 0) {
		throw SocketException("Close socket error !");
	}
}

int16_t DatagramSocket::GetHandler()
{
	return _fd;
}

sockaddr_in DatagramSocket::LocalAddress()
{
	return _lsock;
}

sockaddr_in DatagramSocket::RemoteAddress()
{
	return _server_sock;
}

InetAddress * DatagramSocket::GetInetAddress()
{
	return _address;
}

uint16_t DatagramSocket::GetLocalPort()
{
	return ntohs(_lsock.sin_port);
}

uint16_t DatagramSocket::GetPort()
{
	return ntohs(_server_sock.sin_port);
}

uint64_t DatagramSocket::GetSentBytes()
{
	if (_sent_bytes != 0) {
		return _sent_bytes;
	} else {
		return _buffer->GetSentBytes();
	}
}

uint64_t DatagramSocket::GetReceiveBytes()
{
	if (_receive_bytes != 0) {
		return _receive_bytes;
	} else {
		return _buffer->GetReceiveBytes();
	}
}

SocketOption * DatagramSocket::GetSocketOption()
{
	return new SocketOption(_fd, SOCK_UDP);
}

std::string DatagramSocket::what()
{
	char port[20];
    
	sprintf(port, "%u", GetPort());

	return (GetInetAddress()->GetHostName() + ":" + port);
}

};
